@@ -14,19 +14,16 @@
  * Lesser General Public License for more details.
  *
  * You should have received a copy of the GNU Lesser General Public
- * License along with this library; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+ * License along with this library; if not, see <http://www.gnu.org/licenses/>.
  */
 
 #include <stdint.h>
 #include <stdlib.h>
 #include <stdio.h>
 
-#include "qemu.h"
 #include "cpu.h"
 #include "exec-all.h"
 
-#if !defined (CONFIG_USER_ONLY)
 /* Shared handlers */
 static void pal_reset (CPUState *env);
 /* Console handlers */
@@ -55,8 +52,8 @@ pal_handler_t pal_handlers[] = {
 };
 
 #if 0
-/* One must explicitely check that the TB is valid and the FOE bit is reset */
-static void update_itb ()
+/* One must explicitly check that the TB is valid and the FOE bit is reset */
+static void update_itb (void)
 {
     /* This writes into a temp register, not the actual one */
     mtpr(TB_TAG);
@@ -65,7 +62,7 @@ static void update_itb ()
     mtpr(ITB_PTE);
 }
 
-static void update_dtb ();
+static void update_dtb (void);
 {
     mtpr(TB_CTL);
     /* This write into a temp register, not the actual one */
@@ -82,9 +79,7 @@ static void pal_reset (CPUState *env)
 static void do_swappal (CPUState *env, uint64_t palid)
 {
     pal_handler_t *pal_handler;
-    int status;
 
-    status = 0;
     switch (palid) {
     case 0 ... 2:
         pal_handler = &pal_handlers[palid];
@@ -998,22 +993,22 @@ int cpu_ppc_handle_mmu_fault (CPUState *env, uint32_t address, int rw,
     uint64_t physical, page_size, end;
     int prot, zbits, ret;
 
-    if (env->user_mode_only) {
-        ret = 2;
-    } else {
-        ret = virtual_to_physical(env, &physical, &zbits, &prot,
-                                  address, mmu_idx, rw);
-    }
+    ret = virtual_to_physical(env, &physical, &zbits, &prot,
+                              address, mmu_idx, rw);
+
     switch (ret) {
     case 0:
         /* No fault */
         page_size = 1ULL << zbits;
         address &= ~(page_size - 1);
+        /* FIXME: page_size should probably be passed to tlb_set_page,
+           and this loop removed.   */
         for (end = physical + page_size; physical < end; physical += 0x1000) {
-            ret = tlb_set_page(env, address, physical, prot,
-                               mmu_idx, is_softmmu);
+            tlb_set_page(env, address, physical, prot, mmu_idx,
+                         TARGET_PAGE_SIZE);
             address += 0x1000;
         }
+        ret = 0;
         break;
 #if 0
     case 1:
@@ -1051,51 +1046,3 @@ int cpu_ppc_handle_mmu_fault (CPUState *env, uint32_t address, int rw,
     return ret;
 }
 #endif
-
-#else /* !defined (CONFIG_USER_ONLY) */
-void pal_init (CPUState *env)
-{
-}
-
-void call_pal (CPUState *env, int palcode)
-{
-    target_ulong ret;
-
-    printf("%s: palcode %02x\n", __func__, palcode);
-    if (logfile != NULL)
-        fprintf(logfile, "%s: palcode %02x\n", __func__, palcode);
-    switch (palcode) {
-    case 0x83:
-        /* CALLSYS */
-        printf("CALLSYS n " TARGET_FMT_ld "\n", env->ir[0]);
-        if (logfile != NULL)
-            fprintf(logfile, "CALLSYS n " TARGET_FMT_ld "\n", env->ir[0]);
-        ret = do_syscall(env, env->ir[IR_V0], env->ir[IR_A0], env->ir[IR_A1],
-                         env->ir[IR_A2], env->ir[IR_A3], env->ir[IR_A4],
-                         env->ir[IR_A5]);
-        env->ir[IR_A3] = ret;
-        if (ret > (target_ulong)(-515)) {
-            env->ir[IR_V0] = 1;
-        } else {
-            env->ir[IR_V0] = 0;
-        }
-        break;
-    case 0x9E:
-        /* RDUNIQUE */
-        env->ir[IR_V0] = env->unique;
-        printf("RDUNIQUE: " TARGET_FMT_lx "\n", env->unique);
-        break;
-    case 0x9F:
-        /* WRUNIQUE */
-        env->unique = env->ir[IR_A0];
-        printf("WRUNIQUE: " TARGET_FMT_lx "\n", env->unique);
-        break;
-    default:
-        printf("%s: unhandled palcode %02x\n", __func__, palcode);
-        if (logfile != NULL)
-            fprintf(logfile, "%s: unhandled palcode %02x\n",
-                    __func__, palcode);
-        exit(1);
-    }
-}
-#endif
