@@ -57,8 +57,8 @@
 // functions
 int i8042_flush(void);
 int i8042_command(int command, u8 *param);
-int kbd_command(int command, u8 *param);
-int aux_command(int command, u8 *param);
+int ps2_kbd_command(int command, u8 *param);
+int ps2_mouse_command(int command, u8 *param);
 void ps2port_setup(void);
 
 #endif // ps2port.h
